#include "ggi_widget.h"
#include "ggi_widget_internal.h"

ggi_widget_t ggiWidgetSingleChildGet(ggi_widget_t widget,int childnum) {

	if (childnum>0) return NULL; /* Can only handle a single child */
	return (ggi_widget_t) widget->children;
}

int ggiWidgetSingleChildLink(ggi_widget_t widget,int childnum,ggi_widget_t newchild,...) {

	if (childnum>0) return 1; /* Can only handle a single child */
	if (widget->children)	 return 1;	/* dito */
	widget->children=newchild;
	newchild->parent=widget;
	widget->childchanged(widget,newchild);
	return 0;
}

ggi_widget_t ggiWidgetSingleChildUnlink(ggi_widget_t widget,int childnum) {

	ggi_widget_t rc;

	if (childnum>0) return NULL; /* Can only handle a single child */
	rc=widget->children;
	if (!rc) return NULL;

	((ggi_widget_t) widget->children)->parent=NULL;
	widget->children=NULL;
	widget->childchanged(widget,NULL);
	return rc;
}

ggi_widget_t ggiWidgetChildlistGet(ggi_widget_t widget,int childnum) {

	int x;
	ggi_widget_childlist_t current;
	ggi_widget_childlist_t *rootptr=(ggi_widget_childlist_t *)&widget->children;

	for(x=0,current=*rootptr;current;current=current->next,x++) {
		if (x==childnum||
		    (LAST_CHILD==childnum&&current->next==NULL))
		    	return current->child;
	}
	return NULL;
}

int ggiWidgetChildlistLink(ggi_widget_t widget,int childnum,ggi_widget_t newchild,...) {

	int x;
	ggi_widget_childlist_t current,newcl;
	ggi_widget_childlist_t *insptr;
	ggi_widget_childlist_t *rootptr=(ggi_widget_childlist_t *)&widget->children;

	current=*(insptr=rootptr);

	for(x=0;current;current=*(insptr=&current->next),x++) {
		if (x==childnum)
		    	break;
	}
	if (!current&&childnum!=LAST_CHILD)
		return 1;

	newcl=malloc(sizeof(*newcl));	
	if (!newcl)
		return 2;

	newcl->next=current;
	newcl->child=newchild;
	*insptr=newcl;
	newchild->parent=widget;
	widget->childchanged(widget,newchild);
	return 0;
}

ggi_widget_t ggiWidgetChildlistUnlink(ggi_widget_t widget,int childnum) {

	int x;
	ggi_widget_childlist_t current;
	ggi_widget_childlist_t *insptr;
	ggi_widget_t rc=NULL;
	ggi_widget_childlist_t *rootptr=(ggi_widget_childlist_t *)&widget->children;

	current=*(insptr=rootptr);

	for(x=0;current;current=*(insptr=&current->next),x++) {
		if (x==childnum||
		    (LAST_CHILD==childnum&&current->next==NULL))
		    	break;
	}
	if (!current)
		return NULL;

	rc=current->child;
	current->child->parent=NULL;
	*insptr=current->next;
	free(current);
	widget->childchanged(widget,NULL);

	return rc;
}



int ggiWidgetRedrawChildren(ggi_widget_t self) {

	ggi_widget_t curchild;
	int x,rc,currc;

	rc=0;
	for(x=0;((curchild=self->getchild(self,x)));x++) {
		/* The parent is expected to flush - more efficient. */
		currc=curchild->redraw(curchild,0);
		if (currc>rc) rc=currc;
	}
	return rc;
}

int ggiWidgetDestroyChildren(ggi_widget_t self) {

	ggi_widget_t curchild;
	int rc;

	rc=0;
	while((curchild=self->unlinkchild(self,0))) {
		rc|=curchild->destroy(curchild);
	}
	return rc;
}
