#include <math.h>
#include <stdarg.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_internal.h"
#include "ggi_widget_highlevel.h"

#define MAXNUMLEN 16

struct dialdata {

	ggi_widget_size_t drawpos,drawsize;
	int doopt;
	int diameter;
	double minangle,maxangle;
	double minvalue,maxvalue;
	int    majorticks,minorticks;
	double smallstep,bigstep;
	int    total,postdot;
	int    is_inside,dragging;
	double currvalue;
	char *name;
	char currvaluestring[MAXNUMLEN+1];
	ggi_widget_t	valindi;
};

static int dial_getsize(ggi_widget_t self) {

	ggi_widget_t curchild;
	
	self->min.x=self->min.y=0;
	self->opt.x=self->opt.y=0;
	
	if ((curchild=self->getchild(self,0))) {
		self->min=curchild->min;
		self->opt=curchild->opt;
	}
	return 0;
}

static int ggiWidgetDialRedraw(ggi_widget_t self,int flush) {

	struct dialdata *mine=self->priv;
	int rc;

	ggiWidgetDrawBorder(self,&mine->drawpos,&mine->drawsize);
	rc=ggiWidgetRedrawChildren(self);
	if (flush)
		ggiFlushRegion(self->vis,self->pos.x,self->pos.y,self->size.x,self->size.y);
	return rc;
}

static int ggiWidgetDialVischanged(ggi_widget_t self) {

	struct dialdata *mine=self->priv;
	ggi_widget_t curchild;
	ggiWidgetGetPlacement(self,&mine->drawpos,&mine->drawsize,&mine->doopt,1,1);

	if ((curchild=self->getchild(self,0))) {
		curchild->vis=self->vis;
		curchild->pos.x =self->pos.x+mine->drawpos.x;
		curchild->pos.y =self->pos.y+mine->drawpos.y;
		curchild->size.x=mine->drawsize.x;
		curchild->size.y=mine->drawsize.y;
		curchild->vischanged(curchild);
	}
	return 0;
}

static int ggiWidgetDialChildchanged(ggi_widget_t self,ggi_widget_t child) {

	dial_getsize(self);	/* Recalc my size */
	
	if (self->parent) self->parent->childchanged(self->parent,self);
	else              self->vischanged(self);

	return 0;
}

static void cb_new_value(ggi_widget_t self,int updateodo) {

	struct dialdata *mine=self->priv;

	if (mine->currvalue<mine->minvalue)
		mine->currvalue=mine->minvalue;
	if (mine->currvalue>mine->maxvalue)
		mine->currvalue=mine->maxvalue;
	
	if (self->statevar)
		*((double *)self->statevar)=mine->currvalue;

	snprintf(mine->currvaluestring,sizeof(mine->currvaluestring),"%*.*f",mine->total,mine->postdot,mine->currvalue);
	mine->valindi->childchanged(mine->valindi,NULL);
	if (updateodo) {
		self->redraw(self,1);
	}
	if (self->callback) {
		self->callback(GWT_CB_STATECHANGE,self);
	}
}

static void cb_small_fw(ggiWidgetCallbackType CBTYPE,ggi_widget_t self) {
	ggi_widget_t parent=self->callbackpriv;
	struct dialdata *mine=parent->priv;

	if (CBTYPE!=GWT_CB_ACTIVATE) return;
	mine->currvalue+=mine->smallstep;
	cb_new_value(parent,1);
}

static void cb_big_fw(ggiWidgetCallbackType CBTYPE,ggi_widget_t self) {
	ggi_widget_t parent=self->callbackpriv;
	struct dialdata *mine=parent->priv;

	if (CBTYPE!=GWT_CB_ACTIVATE) return;
	mine->currvalue+=mine->bigstep;
	cb_new_value(parent,1);
}

static void cb_small_back(ggiWidgetCallbackType CBTYPE,ggi_widget_t self) {
	ggi_widget_t parent=self->callbackpriv;
	struct dialdata *mine=parent->priv;

	if (CBTYPE!=GWT_CB_ACTIVATE) return;
	mine->currvalue-=mine->smallstep;
	cb_new_value(parent,1);
}

static void cb_big_back(ggiWidgetCallbackType CBTYPE,ggi_widget_t self) {
	ggi_widget_t parent=self->callbackpriv;
	struct dialdata *mine=parent->priv;

	if (CBTYPE!=GWT_CB_ACTIVATE) return;
	mine->currvalue-=mine->bigstep;
	cb_new_value(parent,1);
}

static void cb_odo(ggiWidgetCallbackType CBTYPE,ggi_widget_t self) {
	ggi_widget_t parent=self->callbackpriv;
	if (CBTYPE!=GWT_CB_STATECHANGE) return;
	cb_new_value(parent,1);
}


ggi_widget_t ggiWidgetCreateDial(va_list ap) {

	ggi_widget_t newid,label,numlabel,smallfw,smallback,bigfw,bigback,odo;
	ggi_widget_t allstack,bottomline,fwstack,backstack,labstack;
	struct dialdata *newdial;

	static struct ggiWidgetBitmap smallbackbit={
		{6,5},"\x10\x20\x40\x20\x10"
	};
	static struct ggiWidgetBitmap smallfwbit={
		{6,5},"\x20\x10\x08\x10\x20"
	};
	static struct ggiWidgetBitmap bigbackbit={
		{6,5},"\x24\x48\x90\x48\x24"
	};
	static struct ggiWidgetBitmap bigfwbit={
		{6,5},"\x90\x48\x24\x48\x90"
	};

	newid=ggiWidgetAllocate();
	if (!newid) return NULL;

	if (NULL==(newdial=malloc(sizeof(*newdial)))) {
		free(newid);
		return NULL;
	}
	newdial->is_inside=0;
	newdial->dragging =0;
	newid->priv=newdial;

	newdial->name    =va_arg(ap,char *);
	newdial->diameter=va_arg(ap,int);
	newdial->minangle=va_arg(ap,double);
	newdial->maxangle=va_arg(ap,double);
	newdial->minvalue=va_arg(ap,double);
	newdial->currvalue=newdial->minvalue;
	newdial->maxvalue=va_arg(ap,double);
	newdial->majorticks=va_arg(ap,int);
	newdial->minorticks=va_arg(ap,int);
	newdial->smallstep =va_arg(ap,double);
	newdial->bigstep   =va_arg(ap,double);
	newdial->total     =va_arg(ap,int);
	newdial->postdot   =va_arg(ap,int);
	if (newdial->total>MAXNUMLEN) newdial->total=MAXNUMLEN;

	newid->max.x=newid->max.y=-1;

	newid->redraw    =ggiWidgetDialRedraw;
	newid->vischanged=ggiWidgetDialVischanged;
	newid->childchanged=ggiWidgetDialChildchanged;

	newid->getchild   =ggiWidgetSingleChildGet;
	newid->linkchild  =ggiWidgetSingleChildLink;
	newid->unlinkchild=ggiWidgetSingleChildUnlink;

	smallfw  =ggiWidgetMakeBitmapbutton(&smallfwbit);
	smallfw->callbackpriv=newid;
	smallfw->callback    =cb_small_fw;

	bigfw    =ggiWidgetMakeBitmapbutton(&bigfwbit);
	bigfw  ->callbackpriv=newid;
	bigfw  ->callback    =cb_big_fw;

	fwstack  =ggiWidgetCreateWidget(GWT_CONTAINER_STACK,0,0,smallfw,bigfw,NULL);

	smallback=ggiWidgetMakeBitmapbutton(&smallbackbit);
	smallback->callbackpriv=newid;
	smallback->callback    =cb_small_back;

	bigback  =ggiWidgetMakeBitmapbutton(&bigbackbit);
	bigback->callbackpriv=newid;
	bigback->callback    =cb_big_back;
	backstack=ggiWidgetCreateWidget(GWT_CONTAINER_STACK,0,0,smallback,bigback,NULL);

	label    =ggiWidgetCreateWidget(GWT_LABEL,newdial->name);
	numlabel =ggiWidgetCreateWidget(GWT_LABEL,NULL);
	numlabel->statevar=&newdial->currvaluestring;
	newdial->valindi=numlabel;
	labstack =ggiWidgetCreateWidget(GWT_CONTAINER_STACK,1,2,label,numlabel,NULL);

	bottomline=ggiWidgetCreateWidget(GWT_CONTAINER_LINE,0,0,backstack,labstack,fwstack,NULL);

	odo      =ggiWidgetCreateWidget(GWT_ODOMETER,newdial->diameter,newdial->minangle,
					newdial->maxangle,newdial->minvalue,newdial->maxvalue,
					newdial->majorticks,newdial->minorticks);
	odo->statevar    =&newdial->currvalue;
	odo->callbackpriv=newid;
	odo->callback    =cb_odo;

	if (newdial->diameter>=26) {
		allstack =ggiWidgetCreateWidget(GWT_CONTAINER_STACK,0,1,odo,bottomline,NULL);
	} else {
		allstack =ggiWidgetCreateWidget(GWT_CONTAINER_LINE,1,0,odo,bottomline,NULL);
	}

	newid->linkchild(newid,0,allstack);
	cb_new_value(newid,0);
	dial_getsize(newid);

	return newid;
}
