#ifndef _GGI_WIDGET_H
#define _GGI_WIDGET_H

#include <ggi/ggi.h>

#define LAST_CHILD (-1)

typedef enum {
	GWT_CB_ENTER=1,
	GWT_CB_LEAVE=2,
	GWT_CB_GET_FOCUS=4,
	GWT_CB_RELEASE_FOCUS=8,
	GWT_CB_STATECHANGE=16,
	GWT_CB_ACTIVATE=32
} ggiWidgetCallbackType;

typedef enum {
	GWT_GRAV_CENTER=0,
	GWT_GRAV_NORTH=1,
	GWT_GRAV_WEST =2,
	GWT_GRAV_SOUTH=4,
	GWT_GRAV_EAST =8,
	GWT_GRAV_STRETCH_X=GWT_GRAV_WEST |GWT_GRAV_EAST,
	GWT_GRAV_STRETCH_Y=GWT_GRAV_NORTH|GWT_GRAV_SOUTH,
	GWT_GRAV_STRETCH  =GWT_GRAV_STRETCH_X|GWT_GRAV_STRETCH_Y
} ggiWidgetGravityType;

typedef enum { 
	GWT_LABEL,
	GWT_BITMAP,
	GWT_PIXMAP,
	GWT_FRAME,
	GWT_BUTTON,
	GWT_CHECKBOX,
	GWT_OPTIONCYCLER,
	GWT_TEXT,
	GWT_TEXTAREA,
	GWT_CHOICELIST,
	GWT_ODOMETER,
	GWT_DIAL,
	GWT_CONTAINER_LINE,
	GWT_CONTAINER_STACK,
	GWT_CONTAINER_GRID,
	GWT_CONTAINER_CARD,
	GWT_LASTTYPE
} ggiWidgetType;


typedef struct ggiWidgetSize {
	short int x,y;
} ggi_widget_size_t;

typedef struct ggiWidgetBitmap {
	ggi_widget_size_t size;
	unsigned char *data;
} *ggi_widget_bitmap_t;

struct ggiWidgetInputStatus;

typedef struct ggiWidgetColorSet {

	ggi_color background,background_light,background_dark,background_empty;
	ggi_color text,texteditbackground;
	ggi_color odometer;

} ggi_widget_color_set;

extern ggi_widget_color_set ggi_widget_default_colorset;

typedef struct ggiWidget {

	struct ggiWidget *parent;
	ggi_widget_color_set *colorset;

	ggi_widget_size_t min,opt,max;
	
	ggiWidgetGravityType gravity;
	int	entered,has_focus,greyed_out;
	
	ggi_visual_t	vis;
	struct ggiWidgetSize pos,size;

	uint32		  hotkey;
	void             *statevar;
	void            (*callback)(ggiWidgetCallbackType CBTYPE,struct ggiWidget *self);
	void  		 *callbackpriv;

	int 		  (*destroy)     (struct ggiWidget *self);
	int 		  (*redraw)      (struct ggiWidget *self,int flush);
	int 		  (*childchanged)(struct ggiWidget *self,struct ggiWidget *child);
	int 		  (*vischanged)  (struct ggiWidget *self);

	int 		  (*process_event)(struct ggiWidget *self,gii_event *ev,struct ggiWidgetInputStatus *stat);

	void *children;
	struct ggiWidget *(*getchild)   (struct ggiWidget *self,int childnum);
	int               (*linkchild)  (struct ggiWidget *self,int childnum,struct ggiWidget *child,...);
	struct ggiWidget *(*unlinkchild)(struct ggiWidget *self,int childnum);

	void  *priv;

} ggi_widget_struct;

typedef ggi_widget_struct *ggi_widget_t;


ggi_widget_t ggiWidgetAllocate(void);
ggi_widget_t ggiWidgetCreateWidget(ggiWidgetType type,...);

ggi_widget_color_set *ggiWidgetGetColorset(ggi_widget_t widget);

int ggiWidgetProcessEvent(ggi_visual_t vis,gii_event *ev);
int ggiWidgetAttach(ggi_widget_t widget, ggi_visual_t vis,int x,int y,int w,int h);
int ggiWidgetDetach(ggi_widget_t widget, ggi_visual_t vis);


ggi_widget_t ggiWidgetGetCurrentFocus(ggi_widget_t widget);
ggi_widget_t ggiWidgetUnfocus(ggi_widget_t widget);


#endif /* _GGI_WIDGET_H */
