#include <math.h>
#include <stdarg.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_internal.h"

struct odometerdata {

	ggi_widget_size_t drawpos,drawsize;
	int doopt;
	int diameter;
	double minangle,maxangle;
	double minvalue,maxvalue;
	int    majorticks,minorticks;
	int    is_inside,dragging;
	double currvalue;
};

static int odometer_getsize(struct odometerdata *mine,struct ggiWidgetSize *ws) {

	ws->x=mine->diameter;
	ws->y=mine->diameter;
	return 0;
}

static void ggiWidgetOdometerDrawTick(ggi_visual_t vis,double angle,ggi_widget_size_t *center,int r1,int r2) {

	double sinus,cosinus;
	
	sinus=sin(angle);
	cosinus=cos(angle);
	
	ggiDrawLine(vis,center->x-r1*sinus,center->y+r1*cosinus,
			center->x-r2*sinus,center->y+r2*cosinus);
}


static int ggiWidgetOdometerRedraw(ggi_widget_t self,int flush) {

	struct odometerdata *mine=self->priv;
	ggi_widget_color_set *colset=ggiWidgetGetColorset(self);
	int x,y;
	ggi_widget_size_t center;
	double currangle;
	int radius;

	if (self->statevar) {
		mine->currvalue=*((double *)self->statevar);
	}
	currangle = mine->currvalue-mine->minvalue;
	currangle/=(mine->maxvalue -mine->minvalue);
	if (currangle<0.0) currangle=0.0;
	if (currangle>1.0) currangle=1.0;
	currangle*=(mine->maxangle-mine->minangle);
	currangle+=mine->minangle;

	ggiWidgetDrawBorder(self,&mine->drawpos,&mine->drawsize);

	ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background));
	ggiDrawBox(self->vis,self->pos.x+mine->drawpos.x,self->pos.y+mine->drawpos.y,
			mine->drawsize.x,mine->drawsize.y);

	center.x=self->pos.x+mine->drawpos.x+mine->drawsize.x/2;
	center.y=self->pos.y+mine->drawpos.y+mine->drawsize.y/2;
	radius=mine->diameter/2;
	
	ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->text));
	ggiWidgetOdometerDrawTick(self->vis,mine->minangle,&center,radius,radius/2);
	ggiWidgetOdometerDrawTick(self->vis,mine->maxangle,&center,radius,radius/2);

	for(x=0;x<=mine->majorticks;x++) {
		double ang1,ang2;
		ang1=mine->minangle+(x+1)*(mine->maxangle-mine->minangle)/(mine->majorticks+1);
		ang2=mine->minangle+    x*(mine->maxangle-mine->minangle)/(mine->majorticks+1);
		if (x) ggiWidgetOdometerDrawTick(self->vis,ang2,&center,radius,radius*3/4);
		for(y=1;y<=mine->minorticks;y++) {
			double ang3;
			ang3=ang2+y*(ang1-ang2)/(mine->minorticks+1);
			ggiWidgetOdometerDrawTick(self->vis,ang3,&center,radius,radius*7/8);
		}
	}

	ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->odometer));
	ggiWidgetOdometerDrawTick(self->vis,currangle,&center,radius,-radius/4);

	if (flush)
		ggiFlushRegion(self->vis,self->pos.x,self->pos.y,self->size.x,self->size.y);
	return 0;
}

static double getnewmouseval(struct odometerdata *mine,ggi_widget_size_t *mr) {

	double angle,dmin;
	
	angle=atan2(-mr->x,mr->y);
	while(angle>mine->minangle) angle-=2*M_PI;
	while(angle<mine->minangle) angle+=2*M_PI;
	dmin=(mine->minangle-angle+2*M_PI);
	if (angle>mine->maxangle) { 
		if (dmin<angle-mine->maxangle)
			angle=mine->minangle;
		else
			angle=mine->maxangle;
	}
	dmin=(angle-mine->minangle)/(mine->maxangle-mine->minangle);
	dmin*=(mine->maxvalue-mine->minvalue);
	return dmin+mine->minvalue;
}

static int ggiWidgetOdometerEvent(ggi_widget_t self,gii_event *ev,ggi_widget_inputstatus_t status) {

	struct odometerdata *mine=self->priv;
	int is_inside;
	double bigstep,smallstep,newmouseval;
	ggi_widget_size_t mouserel;
	
	bigstep=(mine->maxvalue-mine->minvalue)/(mine->majorticks+1);
	bigstep/=(mine->minorticks+1);
	smallstep=bigstep/10.0;

	is_inside=(status->mouse.x>=self->pos.x+mine->drawpos.x &&
		   status->mouse.x< self->pos.x+mine->drawpos.x+mine->drawsize.x &&
		   status->mouse.y>=self->pos.y+mine->drawpos.y &&
		   status->mouse.y< self->pos.y+mine->drawpos.y+mine->drawsize.y );

	mouserel.x=status->mouse.x-(self->pos.x+mine->drawpos.x+mine->drawsize.x/2);
	mouserel.y=status->mouse.y-(self->pos.y+mine->drawpos.y+mine->drawsize.y/2);
	newmouseval=getnewmouseval(mine,&mouserel);

	if (is_inside&&!mine->is_inside) {
		if (self->callback) {
			self->callback(GWT_CB_ENTER,self);
		}
	} else if (!is_inside&&mine->is_inside) {
		if (self->callback) {
			self->callback(GWT_CB_LEAVE,self);
		}
	};
	mine->is_inside=is_inside;

	if (self->has_focus) {
		switch(ev->any.type) {
			case evKeyPress:
			case evKeyRepeat:
				switch(ev->key.sym) {
					case GIIK_Left:
						mine->currvalue-=smallstep;
						newvalue:
						if (mine->currvalue<mine->minvalue) {
							mine->currvalue=mine->minvalue;
						}
						if (mine->currvalue>mine->maxvalue) {
							mine->currvalue=mine->maxvalue;
						}
						if (self->statevar) 
							*((double *)self->statevar)=mine->currvalue;
						self->redraw(self,1);
						if (self->callback) {
							self->callback(GWT_CB_STATECHANGE,self);
						}
						return 1;
					case GIIK_Right:
						mine->currvalue+=smallstep;
						goto newvalue;
					case GIIK_Up:
						mine->currvalue+=bigstep;
						goto newvalue;
					case GIIK_Down:
						mine->currvalue-=bigstep;
						goto newvalue;
					case GIIK_Enter:
						if (self->callback) {
							self->callback(GWT_CB_ACTIVATE,self);
						}
						return 1;
				}
				break;
		}
	} 
	switch(ev->any.type) {
		case evKeyPress:
		case evKeyRepeat:
			if (ev->key.sym==self->hotkey) {
				ggiWidgetUnfocus(self);
				self->has_focus=1;
				return 1;
			}
			break;
		case evPtrButtonPress:
			if (is_inside) {
				mine->currvalue=newmouseval;
				if (!self->has_focus) {
					ggiWidgetUnfocus(self);
					self->has_focus=1;
				}
				mine->dragging=1;
				goto newvalue;
			}
			break;
		case evPtrButtonRelease:
			mine->dragging=0;
			break;
		case evPtrRelative:
		case evPtrAbsolute:
			if (mine->dragging) {
				mine->currvalue=newmouseval;
				goto newvalue;
			}
			break;
	}
	return 0;
}


static int ggiWidgetOdometerVischanged(ggi_widget_t self) {

	struct odometerdata *mine=self->priv;
	ggiWidgetGetPlacement(self,&mine->drawpos,&mine->drawsize,&mine->doopt,1,1);
	return 0;
}

ggi_widget_t ggiWidgetCreateOdometer(va_list ap) {

	ggi_widget_t newid;
	struct odometerdata *newodometer;

	newid=ggiWidgetAllocate();
	if (!newid) return NULL;

	if (NULL==(newodometer=malloc(sizeof(*newodometer)))) {
		free(newid);
		return NULL;
	}
	newodometer->is_inside=0;
	newodometer->dragging =0;
	newid->priv=newodometer;

	newodometer->diameter=va_arg(ap,int);
	newodometer->minangle=va_arg(ap,double);
	newodometer->minangle*=M_PI/180.0;
	newodometer->maxangle=va_arg(ap,double);
	newodometer->maxangle*=M_PI/180.0;
	newodometer->minvalue=va_arg(ap,double);
	newodometer->currvalue=newodometer->minvalue;
	newodometer->maxvalue=va_arg(ap,double);
	newodometer->majorticks=va_arg(ap,int);
	newodometer->minorticks=va_arg(ap,int);

	odometer_getsize(newid->priv,&newid->min);
	newid->opt=newid->min;
	newid->max.x=newid->max.y=-1;

	newid->redraw    =ggiWidgetOdometerRedraw;
	newid->vischanged=ggiWidgetOdometerVischanged;
	newid->process_event=ggiWidgetOdometerEvent;

	return newid;
}
